/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: cast.g,v 1.1 2002/08/14 11:59:22 malekith Exp $
 *
 * Cast - C interface to Ast. This stuff is used by parser.y.
 */
 
open Ast;
open Util;

<'a>located make_located('a data, location loc)
{
	return ({a = data, loc = loc});
}

expr make_expr(raw_expr raw, location loc)
{
	return make_located(({raw = raw, expr_type = None}), loc);
}

ref_record make_ref_record(<var_symbol>bindable name)
{
	return ({sym = name, level = -1});
}

vardef make_vardef(ty var_type, <var_symbol>bindable name)
{
	return ({var_type = var_type, var_name = name});
}

<'a>field make_field(ty fld_type, <'a>bindable fld_symbol, location loc)
{
	return ({fld_symbol = fld_symbol, fld_type = fld_type, fld_loc = loc});
}

nl_decl make_tydef(id name, <id>List::t tyvars, ty typ, location l)
{
	return TyDef[({module = "", 
		       name = name, 
		       stype = typ, 
		       a = ({tysym_tyvars = tyvars}),
		       sloc = l})];
}

<id>List::t make_tyvars(ty ret, <vardef>List::t args) 
{
	let t = ({ret_type = ret, 
           	  parms = List::map(fun *[ty, <id>option] (vardef x) 
		  			([x.var_type, None]), 
				    args),
           	  tyvars = List::Nil}) in
	return Bind::simple_fv(Func[t]);
}
  
nl_decl make_fn_decl(vardef header, string exportas, stmt body)
{
	return FuncDecl[({l = ({lambda_type = header.var_type,
			  	lambda_body = body,
			  	lambda_closure_vars = List::Nil,
			  	lambda_parms = List::Nil}),
			  fn_sym = header.var_name,
			  fn_export_as = exportas})];
}


ty bare_fn_type(ty ret, <*[ty, <id>option]>List::t args)
{
	let t = ({ret_type = ret, 
		  parms = args, 
		  tyvars = List::Nil}) in
	let fv = Bind::simple_fv(Func[t]) in {
		t.tyvars = fv;
		return Func[t];
	}
}

stmt make_vardefs(ty t, <*[<var_symbol>bindable, stmt]>List::t vars)
{
	switch vars {
	case List::Cons[[sym, ini], xs]:
		let def = make_located(VarDef[make_vardef(t, sym)], ini.loc) in
		let p = make_located(Pair[def, ini], ini.loc) in {
			switch xs {
			case List::Nil:
				return p;
			case _:
				return make_located(
					Pair[p, make_vardefs(t, xs)], ini.loc);
			}
		}
	case List::Nil:
		raise Failure["make_vardefs"];
	}
}

nl_expr make_lambda(ty t, stmt body)
{
	return ({raw = 
		Lambda[({lambda_type = t,
			 lambda_body = body,
			 lambda_closure_vars = List::Nil,
			 lambda_parms = List::Nil})],
		 expr_type = None});
}

expr make_ref(<var_symbol>bindable name, location l)
{
	return make_expr(Ref[make_ref_record(name)], l);
}

nl_stmt make_lambda_assign(vardef v, stmt body)
{
	// someday-TODO: this ain't best possible location
	location l = body.loc;
	stmt def = make_located(VarDef[v], l);
	expr lambda = make_located(make_lambda(v.var_type, body), l);
	expr ini_ex = make_expr(SimpleBinOp[Assign, make_ref(v.var_name, l),
					            lambda], l);
	stmt ini = make_located(Expr[ini_ex], l);
	return Pair[def, ini];
}

nl_stmt make_break(break_kind kind, <li_symbol>bindable sym)
{
	return Break[({kind = kind, sym = sym})];
}

nl_stmt make_loop(string label, stmt body, stmt inc)
{
	return Loop[({ll_sym = SimpleName[label], 
		      ll_body = body, 
		      ll_inc = inc})];
}

namedexpr make_namedexpr(<var_symbol>bindable sym, expr ex)
{
	return ({nx_symbol = sym, nx_expr = ex});
}

named_pattern make_named_pattern(<var_symbol>bindable sym, pattern pat)
{
	return ({np_symbol = sym, np_pattern = pat});
}

ty make_named_type(<ty_symbol>bindable tydef, <ty>List::t parms)
{
	return NamedType[({parms = parms, tydef = tydef})];
}

try_record make_try_record(stmt body, <*[pattern, stmt]>List::t try_with,
			   stmt tfinally)
{
	return ({try_body = body, 
		 try_with = try_with, 
		 try_finally = tfinally});
}

expr make_int_const(string s, location l)
{
	try {
		Int64::t v = Int64::of_string(s);
		return make_expr(IntConst[v], l);
	} with {
	case Invalid_argument[_]:
		errorl(l, "invalid integer literal: " + s);
		return make_expr(IntConst[Int64::one], l);
	}
}

expr make_float_const(string s, location l)
{
	try {
		float v = atof(s);
		return make_expr(FloatConst[v], l);
	} with {
	case Invalid_argument[_]:
		errorl(l, "invalid floating point literal: " + s);
		return make_expr(FloatConst[atof("1.0")], l);
	}
}

void raise_located(location l, string msg)
{
	errorl(l, msg);
	raise Choke;
}

lambda_record make_section_record(stmt s)
{
	return ({lambda_body = s,
		 lambda_type = Func[({ret_type = Void,
		 		      parms = List::Nil,
				      tyvars = List::Nil})],
		 lambda_closure_vars = List::Nil,
		 lambda_parms = List::Nil});
}

sini_expr make_sini_expr(<sf_symbol>bindable sym, expr ex)
{
	return ({sx_symbol = sym, sx_expr = ex});
}

ty make_tyvar(id n)
{
	return TyVar[n];
}
