/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: chini.g,v 1.1 2002/08/16 12:17:30 malekith Exp $
 *
 * Chini - check if initialized.
 */

open Ast;
open Util;
open Type;

open Chini;

/* 
   This structure is returned by all chini_* functions. It contains list
   of initialized symbols along with information under what circumstances
   they might be considered initialized. 

   ,,Circumstances'' above are related to loops. This is because symbol
   might be initialized after conditional `break FOO' statement, and should not
   be considered initialized when we finish processing FOO loop. Therefore
   loops are assigned numbers, and so symbols returned in this structure are.

   try { } with { } blocks are treated similarly, as they can be exited
   with raise statement.
   
   Example:

	foo: while (...) {
		x = 1;
		bar: while (true) {
			y = 1;
			if (...)
				break bar;
			z = 1;
			if (...)
				break foo;
			u = 1;
		}
		v = 1;
	}

   Loop foo has number 1, loop bar has number 2. TODO: finish it.
   
 */
struct ini_lst {
	// if syms[s] == N, then s will be initialized before leaving loop 
	// level N+1. Function has loop_level == 0
	<var_symbol, int>Ptrhash::t syms;

	// Lowest level we *will* jump out after executing stmts in
	// question. 0 for function-level (i.e. return stmt),
	// 1 for 1st level loop and so on. -1 if we don't know for sure
	// if we'll jump anywhere.
	int will_break_at;

	// can_break_at[N] is true iff stmt can break at given loop level
	Bitvector::t can_break_at;
}

struct CTX {
	// max(try_level, loop_level)
	int level;
	int try_level;
	int loop_level;

	// maximal loop level we can terminate at
	int can_break_at;

	// initialized symbols, along with loop_levels
	<var_symbol, int>Ptrhash::t initialized;
}

CTX copy_context(CTX c)
{
	return ({ level = c.level,
		  try_level = c.try_level,
		  loop_level = c.loop_level,
		  can_break_at = c.can_break_at,
		  initialized = Ptrhash::copy(c.initialized) });
}

ini_lst nothing_initialized()
{
	return ({ will_break_at = -1, syms = Ptrhash::create(5),
	 	  can_break_at = Bitvector::make(1, false)});
}

void update_break_ctx(CTX c, ini_lst lst)
{
	int i;

	for (i = c.can_break_at + 1; 
	     i < Bitvector::length(lst.can_break_at);
	     i++)
	     	if (Bitvector::get(lst.can_break_at, i))
			c.can_break_at = i;
}

void update_ctx(CTX c, ini_lst lst)
{
	void add(var_symbol s, int lev)
	{
//dbg("uctx: " + symbol_name(s) + " @ " + itoa(Hacks::ptr_of(c.initialized)) + "\n");
		if (Ptrhash::mem(c.initialized, s)) {
			let oldlev = Ptrhash::find(c.initialized, s) in
			if (oldlev > lev) {
				Ptrhash::remove(c.initialized, s);
				Ptrhash::add(c.initialized, s, lev);
			}
		} else {
			Ptrhash::add(c.initialized, s, lev);
		}
	}
	
	Ptrhash::iter(lst.syms, add);
	
	update_break_ctx(c, lst);
}

// Marge [new] into [base], return updated [base] (in-place).
// In [branch] mode it roughly does [(base & new)], in non-branch 
// [(base | new)].
//
// Non-branch mode: if variable is in both [base] and [new], but in [new] it
// is at the lower level, it's moved to that level in [base]. If variable is
// only in [base] it is left were it was. If variable is only in [new], it's
// added to [base] at the same level.
//
// Branch-mode: If variable is only in one of [new] and [base], it's 
// discarded. Otherwise it's placed at upper level of the two. However
// if [new.will_break_at != -1] we do not touch [base]'s
// levels >= [new.will_break_at].

ini_lst merge(ini_lst base, ini_lst new)
{
	<var_symbol, int>Ptrhash::t ret = 
		Ptrhash::create(Ptrhash::length(base.syms) + 
			 	Ptrhash::length(new.syms));

	void add(var_symbol s, int lev)
	{
		try {
			int new_lev = Ptrhash::find(new.syms, s);
			if (new_lev < lev)
				lev = new_lev;
			Ptrhash::remove(new.syms, s);
		} with {
		case Not_found:
	    		Ptrhash::add(ret, s, lev);
		}
	}

	Ptrhash::iter(base.syms, add);

	void add2(var_symbol s, int lev)
	{
		Ptrhash::add(ret, s, lev); 
	}

	Ptrhash::iter(new.syms, add2);

	base.syms = ret;

	return base;
}

ini_lst merge_branch(ini_lst base, ini_lst new)
{
	if (base.will_break_at == 0 && new.will_break_at != 0)
		return merge_branch(new, base);
		
	if (new.will_break_at == -1)
		base.will_break_at = -1;
	if (base.will_break_at != -1 &&
	    new.will_break_at > base.will_break_at)
		base.will_break_at = new.will_break_at;

	<var_symbol, int>Ptrhash::t ret = 
		Ptrhash::create(Ptrhash::length(base.syms) + 
			 	Ptrhash::length(new.syms));

	void add(var_symbol s, int lev)
	{
		if (new.will_break_at != -1 && lev >= new.will_break_at)
			Ptrhash::add(ret, s, lev);
		else
			try {
				let new_lev = Ptrhash::find(new.syms, s) in
				if (new_lev > lev)
					lev = new_lev;
	    			Ptrhash::add(ret, s, lev);
			} with { 
			case Not_found:
				if (new.will_break_at != -1)
					Ptrhash::add(ret, s, 
						     new.will_break_at);
			}
	}

	Ptrhash::iter(base.syms, add);

	base.syms = ret;
	
	return base;
}

void dbg_dump_ini_lst(ini_lst l)
{
	print_string("[ini_lst will_break_at = " + itoa(l.will_break_at) + 
		     ", syms = [ ");
	
	void prt_one(var_symbol s, int lev)
	{
		print_string(symbol_name(s) + ":" + itoa(lev) + " ");
	}

	Ptrhash::iter(l.syms, prt_one);

	print_string("]]\n");
}

ini_lst chini_seq(CTX c, *(ini_lst (CTX, <'a>located)) chini_one, 
		  <<'a>located>List::t data)
{
	ini_lst r = nothing_initialized();
	bool did_warn = false;

	void do_one(<'a>located d)
	{
		let rr = chini_one(c, d) in {
//dbg("one: ");
//dbg_dump_ini_lst(r);
			if (r.will_break_at == -1) {
				r.will_break_at = rr.will_break_at;
			} else {
				if (flags.wdead && !did_warn) {
					warnl(d.loc, "dead code");
					did_warn = true;
				}
			}
			update_ctx(c, rr);
			r = merge(r, rr);
		}
	}

	List::iter(do_one, data);

//dbg("seq: ");
//dbg_dump_ini_lst(r);

	return r;
}

// We're not at level [n], fixup [l] as aproptiate 
ini_lst cut_down(int n, ini_lst l)
{
	<var_symbol, int>Ptrhash::t ret = 
		Ptrhash::create(Ptrhash::length(l.syms));
	
	void add(var_symbol s, int lev)
	{
		if (lev <= n)
			Ptrhash::add(ret, s, lev);
	}
	
	Ptrhash::iter(l.syms, add);
	l.syms = ret;
	
	Bitvector::resize(l.can_break_at, n + 1, false);
	if (l.will_break_at == n + 1)
		l.will_break_at = -1;

	return l;
}

ini_lst chini_branches(CTX c, <stmt>List::t branches)
{
	ini_lst fold(<stmt>List::t br, ini_lst il)
	{
//dbg("branches_got: ");
//dbg_dump_ini_lst(il);

		switch br {
		case List::Cons[s, ss]:
			let cc = copy_context(c) in
			let is = chini_stmt(cc, s) in
			return fold(ss, merge_branch(il, is));
			
		case List::Nil:
			return il;
		}
	}

	let r = nothing_initialized() in {
		r.will_break_at = 0;
		return fold(branches, r);
	}
}

ini_lst chini_symbol(CTX c, var_symbol v)
{
	if (has_property(v, IsInitialized) ||
	    has_property(v, IsUnionField) ||
	    has_property(v, IsInPattern) ||
	    v.a.scope_level <= 0)
		return nothing_initialized();		// ok
		
	try {
		Ptrhash::find(c.initialized, v);
	} with {
	case Not_found:
		error("variable " + symbol_name(v) + 
		      " might be used uninitalized");
	}

	return nothing_initialized();
}

ini_lst chini_fnc(CTX c, lambda_record l, string name)
{
	ini_lst i = chini_stmt(c, l.lambda_body);

	if (!is_void_func_type(l.lambda_type) && i.will_break_at != 0)
		error(name + " can finish without returning a value");
	
	return nothing_initialized();
}

ini_lst chini_assign(CTX c, expr e1, expr e2)
{
	ini_lst i1 = nothing_initialized(), i2;
	<var_symbol>List::t syms = List::Nil;
	
	void collect_symbols(expr e)
	{
		switch e.a.raw {
		case Ref[r]:
			var_symbol s = get_binding(r.sym);
			int lev = c.can_break_at + 1;
			
			if (Ptrhash::mem(i1.syms, s))
				return;
			
			try {
				lev = Ptrhash::find(c.initialized, s);
			} with { case Not_found: skip; }
		
			if (lev > c.can_break_at)
				Ptrhash::add(i1.syms, s, c.can_break_at);
			
		case TupleExpr[l]:
			List::iter(collect_symbols, l);

		case _:
			skip;
		}
	}
	
	collect_symbols(e1);

	// If e2 is lambda expression, we update context first, as there
	// is no way it can be called before this assignment is done. 
	// Otherwise local functions cannot be made self-recursive.
	// Note that it still doesn't allow mutual recursion which,
	// generally speaking, sucks eggs.
	switch e2.a.raw {
	case Lambda[_]:
		update_ctx(c, i1);
		i2 = chini_expr(c, e2);
	case _:
		i2 = chini_expr(c, e2);
		update_ctx(c, i1);
	}

	update_ctx(c, i2);

	return merge(i1, i2);
}

ini_lst chini_nl_expr(CTX c, nl_expr ex)
{
	ini_lst seq(<expr>List::t l)
	{
		return chini_seq(c, chini_expr, l);
	}
	
	switch ex.raw {
	case SimpleBinOp[Assign, e1, e2]:
		return chini_assign(c, e1, e2);
		
	case SimpleBinOp[_, e1, e2]:
		return seq(List::cons2(e1, e2));
		
	case SimpleUnOp[_, e]:
		return chini_expr(c, e);
	
	case Ref[rr]:
		return chini_symbol(c, get_binding(rr.sym));

	case Call[e, el]:
		return seq(List::Cons[e, el]);

	case TupleExpr[el]:
		return seq(el);

	case UnionExpr[u]:
		return chini_expr(c, u.nx_expr);
	
	case Lambda[l]:
		return chini_fnc(c, l, "this lambda-expression");

	case StructExpr[s]:
		return seq(List::map(fun expr (sini_expr se) (se.sx_expr), s));

	case DotExpr[de]:
		return chini_expr(c, de.sx_expr);
	
	// contants arn't really interesting in this context ;)
	case BoolConst[_]:	return nothing_initialized();
	case IntConst[_]: 	return nothing_initialized();
	case FloatConst[_]: 	return nothing_initialized();
	case StringConst[_]:	return nothing_initialized();
	case VoidConst:		return nothing_initialized();
	case NullConst:		return nothing_initialized();
	}
}

ini_lst chini_expr(CTX c, expr ex)
{
	return locate(partial(chini_nl_expr, c), ex);
}

void can_break_at(ini_lst i, int lev)
{
	if (Bitvector::length(i.can_break_at) <= lev)
		Bitvector::resize(i.can_break_at, lev + 1, false);
	Bitvector::set(i.can_break_at, lev, true);
}

ini_lst chini_nl_stmt(CTX c, nl_stmt nlr)
{
	switch nlr {
	case Skip:
		return nothing_initialized();
	
	case Expr[e]:
		return chini_expr(c, e);
	
	case Return[e]:
		let r = chini_expr(c, e) in {
			r.will_break_at = 0;
			return r;
		}

	case VarDef[_]:
		return nothing_initialized();
		
  	case Block[lst]:
		return chini_seq(c, chini_stmt, lst);

	case Pair[s1, s2]:
		return chini_seq(c, chini_stmt, List::cons2(s1, s2));
	
	case Switch[e, clauses]:
		let pats = List::map(fst, clauses) in
		Patrange::check_exhaustiveness(pats);
		
		let exret = chini_expr(c, e) in {
			update_ctx(c, exret);
			let b = List::map(snd, clauses) in
			return merge(chini_branches(c, b), exret);
		}
		
	case Let[assigns, s]:
		List::iter(Patrange::check_exhaustiveness, 
			   List::map(List::cons1, List::map(fst, assigns)));
		let a = chini_seq(c, chini_expr, List::map(snd, assigns)) in
		return merge(chini_stmt(c, s), a);
	
	case If[e, s1, s2]:
		let exret = chini_expr(c, e) in {
			update_ctx(c, exret);
			let b = List::cons2(s1, s2) in
			return merge(chini_branches(c, b), exret);
		}
	
	case Loop[l]:
		CTX lc = copy_context(c);
		
		let lev = lc.level + 1 in {
			lc.level = lev;
			(get_binding(l.ll_sym)).a.loop_level = lev;
			lc.loop_level = lev;
		}
		
		let body_ini = chini_stmt(lc, l.ll_body) in
		let ret = cut_down(c.level, body_ini) in {
			// it might never be executed, so we ignore the result,
			// except for the can_break part
			ini_lst inc = chini_stmt(c, l.ll_inc);
			Bitvector::or(ret.can_break_at, inc.can_break_at);
			return ret;
		}
	
	case Break[b]:
		let r = nothing_initialized() in {
			r.will_break_at = (get_binding(b.sym)).a.loop_level;
			can_break_at(r, r.will_break_at);
			return r;
		}
		
	case BreakIfNot[e]:
		let r = chini_expr(c, e) in {
			switch e.a.raw {
			case BoolConst[false]:
				r.will_break_at = c.loop_level;
			case _:
				skip;
			}

			can_break_at(r, c.loop_level);
			
			return r;
		}
	
	case Open[_]:
		return nothing_initialized();
	
	case Try[tr]:
		CTX tc = copy_context(c);
		tc.try_level = ++tc.level;

		// FIXME: OGCB
		ini_lst body_ret = chini_stmt(tc, tr.try_body);
		bool restore = body_ret.will_break_at == c.level + 1;

		let cut_ret = cut_down(c.level, body_ret) in
		let with_matching = List::map(snd, tr.try_with) in
		let with_ret = chini_branches(c, with_matching) in
		let ret = merge_branch(cut_ret, with_ret) in
		let fc = copy_context(c) in {
//dbg("brwba= " +itoa(body_ret.will_break_at)+"\n");
			if (restore)
			    	ret.will_break_at = with_ret.will_break_at;

			// statements in finally are *always* executed, no
			// matter how deep loop we happen to jump out
			fc.can_break_at = c.can_break_at;
			let fret = chini_stmt(fc, tr.try_finally) in
			return merge(ret, fret);
		}
	
	case Raise[e]:
		let r = chini_expr(c, e) in {
			r.will_break_at = c.try_level;
			can_break_at(r, c.try_level);
			return r;
		}
	}
}

ini_lst chini_stmt(CTX c, stmt s)
{
	let il = locate(partial(chini_nl_stmt, c), s) in {
	
//dbg("stmt: ");
//dbg_dump_ini_lst(il);
	return il;}
}

CTX make_ctx()
{
	return ({level = 0,
		 try_level = 0,
		 loop_level = 0,
		 can_break_at = 0,
		 initialized = Ptrhash::create(13)
	        });
}

void chini_nl_decl(nl_decl d)
{
	switch d {
	case FuncDecl[fn]:
		chini_fnc(make_ctx(), fn.l, qualified_name(fn.fn_sym) + "()");
	
	case SectionDecl[_, l]:
		chini_fnc(make_ctx(), l, "this section");
	
	case TopLevelOpen[_]:		skip;
	case GlobalDecl[_]:		skip;
	case ExternC[_]:		skip;
	case FuncProto[_]:		skip;
	case XCFuncProto[_, _]:		skip;
	case TyDef[_]:			skip;
	case ExceptionDecl[_]:		skip;
	case DummyDecl:			skip;
	}
}

void chini_decl(decl d)
{
	locate(chini_nl_decl, d);
}
